from functools import lru_cache


class IdUtil(object):
    """
    ID util.
    """

    @staticmethod
    @lru_cache(maxsize=32768)
    def hash_string(s):
        """
        Hashes the string.

        :param s: String.
        :return: Hash value.
        """
        hash_value = 0

        if len(s) == 0:
            return hash_value

        for c in s:
            hash_value = ((hash_value << 5) - hash_value) + ord(c)
            hash_value |= 0

        return hash_value



import json

import networkx as nx

from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.graph import Graph
from pybbn.graph.node import BbnNode
from pybbn.graph.variable import Variable


class Dag(Graph):
    """
    Directed acyclic graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        Graph.__init__(self)

    def get_parents(self, id):
        """
        Gets the parent IDs of the specified node.

        :param id: Node id.
        :return: Array of parent ids.
        """
        return [x for x in self.edge_map if id in self.edge_map[x]]

    def get_children(self, node_id):
        """
        Gets the children IDs of the specified node.

        :param node_id: Node id.
        :return: Array of children ids.
        """
        return [x for x in self.edge_map[node_id]]

    def __shouldadd__(self, edge):
        """
        Checks if the specified directed edge should be added.

        :param edge: Directed edge.
        :return: A boolean indicating if the edge should be added.
        """
        if EdgeType.DIRECTED != edge.type:
            return False

        parent = edge.i
        child = edge.j

        if parent.id == child.id:
            return False

        if child.id not in self.edge_map[parent.id] and parent.id not in self.edge_map[child.id]:
            if not PathDetector(self, child.id, parent.id).exists():
                return True

        return False

    def edge_exists(self, id1, id2):
        """
        Checks if a directed edge exists between the specified id. e.g. id1 -> id2

        :param id1: Node id.
        :param id2: Node id.
        :return: A boolean indicating if a directed edge id1 -> id2 exists.
        """
        if id2 in self.edge_map[id1] and id1 not in self.edge_map[id2]:
            return True
        return False

    def to_nx_graph(self):
        """
        Converts this DAG to a NX DiGraph for visualization.

        :return: A tuple, where the first item is the NX DiGraph and the second items are the node labels.
        """
        g = nx.DiGraph()
        labels = []

        for k, node in self.nodes.items():
            g.add_node(node.id)
            t = (node.id, node.variable.name)
            labels.append(t)

        for k, edge in self.edges.items():
            pa = edge.i.id
            ch = edge.j.id
            g.add_edges_from([(pa, ch, {})])

        return g, dict(labels)


class Bbn(Dag):
    """
    BBN.
    """

    def __init__(self):
        """
        Ctor.
        """
        Dag.__init__(self)
        self.parents = {}

    def get_parents_ordered(self, id):
        """
        Gets the IDs of the specified node ordered.

        :param id: ID of node.
        :return: List of parent IDs sorted.
        """
        return sorted(self.parents[id]) if id in self.parents else []

    def __edge_added__(self, edge):
        if edge.j.id not in self.parents:
            self.parents[edge.j.id] = []

        if edge.i.id not in self.parents[edge.j.id]:
            self.parents[edge.j.id].append(edge.i.id)

    def __shouldadd__(self, edge):
        """
        Checks if the specified directed edge should be added.

        :param edge: Directed edge.
        :return: A boolean indicating if the directed edge should be added.
        """
        if isinstance(edge.i, BbnNode) and isinstance(edge.j, BbnNode):
            return True
        return Dag.__shouldadd__(edge)

    @staticmethod
    def to_csv(bbn, path):
        """
        Converts the specified BBN to CSV format.

        :param bbn: BBN.
        :param path: Path to file.
        :return: None.
        """
        with open(path, 'w') as f:
            for node in bbn.get_nodes():
                v = node.variable
                vals = ','.join(v.values)
                probs = ','.join([str(p) for p in node.probs])
                s_node = f'{v.id},{v.name},{vals},|,{probs}'
                f.write(s_node)
                f.write('\n')

            for _, edge in bbn.edges.items():
                t = 'directed' if edge.type == EdgeType.DIRECTED else 'undirected'
                s_edge = f'{edge.i.id},{edge.j.id},{t}'
                f.write(s_edge)
                f.write('\n')

    @staticmethod
    def from_csv(path):
        """
        Converts the BBN in CSV format to a BBN.
        :param path: Path to CSV file.
        :return: BBN.
        """
        with open(path, 'r') as f:
            nodes = {}
            edges = []

            for line in f:
                tokens = line.split(',')
                if 3 == len(tokens):
                    edge = int(tokens[0]), int(tokens[1])
                    edges.append(edge)
                else:
                    tokens = line.split('|')
                    v_part = [item.strip() for item in tokens[0].split(',') if len(item.strip()) > 0]
                    p_part = [item.strip() for item in tokens[1].split(',') if len(item.strip()) > 0]

                    i = int(v_part[0])
                    v = Variable(i, v_part[1], v_part[2:])
                    p = [float(p) for p in p_part]

                    node = BbnNode(v, p)
                    nodes[i] = node

            bbn = Bbn()
            for _, node in nodes.items():
                bbn.add_node(node)
            for edge in edges:
                pa_id, ch_id = edge

                pa = nodes[pa_id]
                ch = nodes[ch_id]

                bbn.add_edge(Edge(pa, ch, EdgeType.DIRECTED))
            return bbn

    @staticmethod
    def to_dict(bbn):
        """
        Gets a JSON serializable dictionary representation.

        :param bbn: BBN.
        :return: Dictionary.
        """
        return {
            'nodes': {n.id: n.to_dict() for n in bbn.get_nodes()},
            'edges': [{'pa': edge.i.id, 'ch': edge.j.id} for _, edge in bbn.edges.items()]
        }

    @staticmethod
    def from_dict(d):
        """
        Creates a BBN from a dictionary (deserialized JSON).

        :param d: Dictionary.
        :return: BBN.
        """

        def get_variable(d):
            return Variable(d['id'], d['name'], d['values'])

        def get_bbn_node(d):
            return BbnNode(get_variable(d['variable']), d['probs'])

        nodes = {k: get_bbn_node(n) for k, n in d['nodes'].items()}
        edges = d['edges']

        bbn = Bbn()

        for k, n in nodes.items():
            bbn.add_node(n)

        for e in edges:
            pa_id = e['pa']
            ch_id = e['ch']

            pa = nodes[pa_id] if pa_id in nodes else nodes[str(pa_id)]
            ch = nodes[ch_id] if ch_id in nodes else nodes[str(ch_id)]

            bbn.add_edge(Edge(pa, ch, EdgeType.DIRECTED))

        return bbn

    @staticmethod
    def to_json(bbn, path):
        """
        Serializes BBN to JSON.

        :param bbn: BBN.
        :param path: Path.
        :return: None.
        """
        s = json.dumps(Bbn.to_dict(bbn), indent=2)
        with open(path, 'w') as f:
            f.write(s)

    @staticmethod
    def from_json(path):
        """
        Deserializes BBN from JSON.

        :param path: Path.
        :return: BBN.
        """
        with open(path, 'r') as f:
            d = json.loads(f.read())
            bbn = Bbn.from_dict(d)
            return bbn


class PathDetector(object):
    """
    Detects path between two nodes.
    """

    def __init__(self, graph, start, stop):
        """
        Ctor.

        :param graph: DAG.
        :param start: Start node id.
        :param stop: Stop node id.
        """
        self.graph = graph
        self.start = start
        self.stop = stop
        self.seen = set()

    def exists(self):
        """
        Checks if a path exists.

        :return: True if a path exists, otherwise, false.
        """
        if self.start == self.stop:
            return True
        else:
            return self.__find__(self.start)

    def __find__(self, i):
        """
        Checks if a path exists from the specified node to the stop node.

        :param i: Node id.
        :return: True if a path exists, otherwise, false.
        """
        children = self.graph.get_children(i)
        if self.stop in children:
            return True

        self.seen.add(i)
        for child in children:
            if child not in self.seen and self.__find__(child):
                return True

        return False


class BbnUtil(object):
    """
    BBN utility.
    """

    @staticmethod
    def get_huang_graph():
        """
        Gets the Huang reference BBN graph.

        :return: BBN.
        """
        a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
        b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
        c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
        d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])
        e = BbnNode(Variable(4, 'e', ['on', 'off']), [0.3, 0.7, 0.6, 0.4])
        f = BbnNode(Variable(5, 'f', ['on', 'off']), [0.01, 0.99, 0.01, 0.99, 0.01, 0.99, 0.99, 0.01])
        g = BbnNode(Variable(6, 'g', ['on', 'off']), [0.8, 0.2, 0.1, 0.9])
        h = BbnNode(Variable(7, 'h', ['on', 'off']), [0.05, 0.95, 0.95, 0.05, 0.95, 0.05, 0.95, 0.05])

        bbn = Bbn() \
            .add_node(a) \
            .add_node(b) \
            .add_node(c) \
            .add_node(d) \
            .add_node(e) \
            .add_node(f) \
            .add_node(g) \
            .add_node(h) \
            .add_edge(Edge(a, b, EdgeType.DIRECTED)) \
            .add_edge(Edge(a, c, EdgeType.DIRECTED)) \
            .add_edge(Edge(b, d, EdgeType.DIRECTED)) \
            .add_edge(Edge(c, e, EdgeType.DIRECTED)) \
            .add_edge(Edge(d, f, EdgeType.DIRECTED)) \
            .add_edge(Edge(e, f, EdgeType.DIRECTED)) \
            .add_edge(Edge(c, g, EdgeType.DIRECTED)) \
            .add_edge(Edge(e, h, EdgeType.DIRECTED)) \
            .add_edge(Edge(g, h, EdgeType.DIRECTED))

        return bbn

    @staticmethod
    def get_simple():
        """
        Gets a simple BBN graph.

        :return: BBN.
        """
        a = BbnNode(Variable(0, 'a', ['on', 'off']), [0.5, 0.5])
        b = BbnNode(Variable(1, 'b', ['on', 'off']), [0.5, 0.5, 0.4, 0.6])
        c = BbnNode(Variable(2, 'c', ['on', 'off']), [0.7, 0.3, 0.2, 0.8])
        d = BbnNode(Variable(3, 'd', ['on', 'off']), [0.9, 0.1, 0.5, 0.5])
        e = BbnNode(Variable(4, 'e', ['on', 'off']), [0.3, 0.7, 0.6, 0.4])
        f = BbnNode(Variable(5, 'f', ['on', 'off']), [0.01, 0.99, 0.01, 0.99, 0.01, 0.99, 0.99, 0.01])

        bbn = Bbn() \
            .add_node(a) \
            .add_node(b) \
            .add_node(c) \
            .add_node(d) \
            .add_node(e) \
            .add_node(f) \
            .add_edge(Edge(a, b, EdgeType.DIRECTED)) \
            .add_edge(Edge(a, c, EdgeType.DIRECTED)) \
            .add_edge(Edge(b, d, EdgeType.DIRECTED)) \
            .add_edge(Edge(c, e, EdgeType.DIRECTED)) \
            .add_edge(Edge(d, f, EdgeType.DIRECTED)) \
            .add_edge(Edge(e, f, EdgeType.DIRECTED))

        return bbn



import bisect
import copy
import itertools
from functools import cmp_to_key

import numpy as np


class SortableNode(object):
    """
    Sortable node.
    """

    def __init__(self, node_id, parent_ids):
        """
        Ctor.

        :param node_id: Node ID.
        :param parent_ids: List of parent IDs.
        """
        self.node_id = node_id
        self.parent_ids = parent_ids


class Table(object):
    """
    Table association parent instantiations with cumulative distributions
    of node values.
    """

    def __init__(self, node, parents=[]):
        """
        Ctor.

        :param node: BBN node.
        :param parents: List of parent BBN nodes.
        """
        self.node = node
        self.parents = sorted(parents, key=lambda n: n.id)
        self.parent_ids = [p.id for p in parents]

        if len(parents) == 0:
            self.probs = np.array([node.probs]).cumsum()
        else:
            cartesian = list(itertools.product(*[node.variable.values for node in self.parents]))
            get_kv = lambda i, v: f'{i}={v}'
            keys = [','.join([get_kv(node.id, val) for node, val in zip(self.parents, values)]) for values in cartesian]
            n = len(node.variable.values)

            probs = [node.probs[i:i + n] for i in range(0, len(node.probs), n)]
            probs = [np.array(p).cumsum() for p in probs]
            self.probs = {k: p for k, p in zip(keys, probs)}

    def get_value(self, prob, sample=None):
        """
        Gets the value associated with the specified probability.

        :param prob: Probability.
        :param sample: Dictionary of variable-value sampled so far.
        :return: Value.
        """
        if not self.has_parents():
            index = bisect.bisect(self.probs, prob)
        else:
            k = ','.join([f'{i}={sample[i]}' for i in self.parent_ids])
            probs = self.probs[k]
            index = bisect.bisect(probs, prob)
        return self.node.variable.values[index]

    def has_parents(self):
        """
        Checks if the node associated with this table has parents.

        :return: Boolean.
        """
        return self.parents is not None and len(self.parents) > 0


class LogicSampler(object):
    """
    Logic sampling with rejection.
    """

    def __init__(self, bbn):
        """
        Ctor.

        :param bbn: BBN.
        """
        self.bbn = bbn
        self.nodes = self.__topological_sort__()
        self.tables = self.__get_tables__()

    def __get_tables__(self):
        """
        Gets a dictionary of node ID to Table.

        :return: Dictionary of node ID to Table.
        """
        tables = [node for node in self.bbn.get_nodes()]
        tables = [(node, self.bbn.get_parents_ordered(node.id)) for node in tables]
        tables = [(node, [self.bbn.get_node(pa_id) for pa_id in parents]) for node, parents in tables]
        tables = {node.id: Table(node, parents) for node, parents in tables}
        return tables

    def __topological_sort__(self):
        """
        Performs topological sort of nodes.

        :return: List of node IDs that is topologically sorted.
        """
        nodes = [SortableNode(node.id, set(self.bbn.get_parents_ordered(node.id))) for node in self.bbn.get_nodes()]
        id_sorter = lambda x, y: -1 if x < y else 1 if x > y else 0
        pa_sorter = lambda x, y: -1 if x.node_id in y.parent_ids else 1 if y.node_id in x.parent_ids else id_sorter(
            x.node_id, y.node_id)
        nodes = sorted(nodes, key=cmp_to_key(pa_sorter))
        nodes = [n.node_id for n in nodes]
        return nodes

    def get_samples(self, evidence={}, n_samples=100, seed=37):
        """
        Gets the samples.

        :param evidence: Evidence. Dictionary. Keys are ids and values are node values.
        :param n_samples: Number of samples.
        :param seed: Seed (default=37).
        :return: Samples.
        """
        np.random.seed(seed)
        n = len(self.nodes)
        samples = []
        while True:
            sample = copy.deepcopy(evidence)

            probs = np.random.uniform(0.0, 1.0, size=n)
            for p, node_id in zip(probs, self.nodes):
                table = self.tables[node_id]
                val = table.get_value(p, sample=sample)

                if node_id in evidence and val != evidence[node_id]:
                    break
                sample[node_id] = val

            if len(sample) == n:
                samples.append(sample)

            if len(samples) == n_samples:
                break

        return samples



import copy

from pybbn.graph.jointree import JoinTreeListener
from pybbn.pptc.initializer import Initializer
from pybbn.pptc.moralizer import Moralizer
from pybbn.pptc.potentialinitializer import PotentialInitializer
from pybbn.pptc.propagator import Propagator
from pybbn.pptc.transformer import Transformer
from pybbn.pptc.triangulator import Triangulator


class InferenceController(JoinTreeListener):
    """
    Inference controller.
    """

    @staticmethod
    def apply(bbn):
        """
        Sets up the specified BBN for probability propagation in tree clusters (PPTC).

        :param bbn: BBN graph.
        :return: Join tree.
        """
        PotentialInitializer.init(bbn)

        ug = Moralizer.moralize(bbn)
        cliques = Triangulator.triangulate(ug)
        join_tree = Transformer.transform(cliques)
        join_tree.parent_info = {node.id: bbn.get_parents_ordered(node.id) for node in bbn.get_nodes()}

        Initializer.initialize(join_tree)
        Propagator.propagate(join_tree)

        join_tree.set_listener(InferenceController())

        return join_tree

    @staticmethod
    def reapply(join_tree, cpts):
        """
        Reapply propagation to join tree with new CPTs. The join tree structure is kept but the BBN node CPTs
        are updated. A new instance/copy of the join tree will be returned.

        :param join_tree: Join tree.
        :param cpts: Dictionary of new CPTs. Keys are id's of nodes and values are new CPTs.
        :return: Join tree.
        """
        jt = copy.deepcopy(join_tree)
        jt.update_bbn_cpts(cpts)
        jt.listener = None
        jt.evidences = dict()

        PotentialInitializer.reinit(jt)
        Initializer.initialize(jt)
        Propagator.propagate(jt)

        jt.set_listener(InferenceController())

        return jt

    @staticmethod
    def apply_from_serde(join_tree):
        """
        Applies propagation to join tree from a deserialzed join tree.

        :param join_tree: Join tree.
        :return: Join tree (the same one passed in).
        """
        join_tree.listener = None
        join_tree.evidences = dict()

        PotentialInitializer.reinit(join_tree)
        Initializer.initialize(join_tree)
        Propagator.propagate(join_tree)

        join_tree.set_listener(InferenceController())

        return join_tree

    def evidence_retracted(self, join_tree):
        """
        Evidence is retracted.

        :param join_tree: Join tree.
        """
        Initializer.initialize(join_tree)
        Propagator.propagate(join_tree)

    def evidence_updated(self, join_tree):
        """
        Evidence is updated.

        :param join_tree: Join tree.
        """
        Propagator.propagate(join_tree)



from pybbn.graph.edge import JtEdge
from pybbn.graph.jointree import JoinTree
from pybbn.graph.node import SepSet


class Transformer(object):
    """
    Transformer. Transforms a list of cliques into a join tree.
    """

    @staticmethod
    def transform(cliques):
        """
        Transforms the cliques into a join tree.

        :param cliques: List of cliques.
        :return: Join tree.
        """
        join_tree = JoinTree()
        for clique in cliques:
            join_tree.add_node(clique)

        sep_sets = Transformer.get_sep_sets(cliques)

        for i in range(len(sep_sets)):
            jtEdge = JtEdge(sep_sets[i])
            join_tree.add_edge(jtEdge)

        return join_tree

    @staticmethod
    def get_sep_sets(cliques):
        """
        Gets all pair-wise separation-sets.

        :param cliques: Array of cliques.
        :return: Array of separation sets sorted descendingly by mass followed by cost (asc) and id (asc).
        """
        sep_sets = []
        size = len(cliques)
        for i in range(size):
            clique_i = cliques[i]
            for j in range(i + 1, size):
                clique_j = cliques[j]
                is_intersection, lhs, rhs, intersection = clique_i.intersects(clique_j)
                if is_intersection:
                    sep_set = SepSet(clique_i, clique_j, lhs, rhs, intersection)
                    sep_sets.append(sep_set)

        return sorted(sep_sets, key=lambda x: (-1 * x.mass, x.cost, x.id))



import math
from collections import namedtuple

import networkx as nx
import numpy as np
from joblib import Parallel, delayed
from networkx.algorithms.dag import topological_sort, is_directed_acyclic_graph
from scipy.stats import multivariate_normal

from pybbn.lg.gaussian import dnorm, dcmvnorm
from pybbn.lg.inference import MvnInference

WORK_UNIT = namedtuple('WORK_UNIT', 'row_id node_id m s dep')
WORK_UNITS = namedtuple('WORK_UNITS', 'row_id units')
WORK_RESULT = namedtuple('WORK_RESULT', 'row_id, prob')


def __get__prob__(X, W):
    if W.dep is None:
        d = [X[W.row_id, W.node_id]]
        return next(dnorm(d, W.m, W.s))
    else:
        d = X[W.row_id, :].reshape(1, -1)
        return next(dcmvnorm(d, W.m, W.s, W.node_id, W.dep))


def __get_probs__(X, W):
    probs = [__get__prob__(X, w) for w in W.units]
    prob = 10 ** np.log10(probs).sum()
    return WORK_RESULT(W.row_id, prob)


def __get_log_prob__(X, W):
    if W.dep is None:
        d = [X[W.row_id, W.node_id]]
        return np.log10(next(dnorm(d, W.m, W.s)))
    else:
        d = X[W.row_id, :].reshape(1, -1)
        return np.log10(next(dcmvnorm(d, W.m, W.s, W.node_id, W.dep)))


def __get_log_probs__(X, W):
    p = np.sum([__get_log_prob__(X, w) for w in W.units])
    return WORK_RESULT(W.row_id, p)


class Dag(object):
    """
    Directed acyclic graph.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.g = nx.DiGraph()

    def add_node(self, node_id, metadata={}):
        """
        Adds a node.

        :param node_id: Node ID.
        :param metadata: Metadata (JSON-like).
        :return: DAG.
        """
        self.g.add_node(node_id, metadata=metadata)
        return self

    def add_edge(self, parent_id, child_id, metadata={}):
        """
        Adds an edge.

        :param parent_id: Parent node ID.
        :param child_id: Child node ID.
        :param metadata: Metadata (JSON-like).
        :return: DAG.
        """
        self.g.add_edges_from([(parent_id, child_id, metadata)])
        if is_directed_acyclic_graph(self.g) is False:
            self.g.remove_edge(parent_id, child_id)
            raise ValueError('adding link from parent {} to child {} will create a cycle'.format(parent_id, child_id))
        return self

    def nodes(self):
        """
        Gets a list of all the node IDs.

        :return: List of node IDs.
        """
        return list(self.g.nodes())

    def edges(self):
        """
        Gets a list of all the edges.

        :return: List of edges.
        """
        return list(self.g.edges())

    def number_of_nodes(self):
        """
        Gets the number of nodes.

        :return: Number of nodes.
        """
        return self.g.number_of_nodes()

    def number_of_edges(self):
        """
        Gets the number of edges.

        :return: Number of edges.
        """
        return self.g.number_of_edges()

    def parents(self, node_id):
        """
        Gets the parents for the specified node ID.

        :param node_id: Node ID.
        :return: List of parent node IDs.
        """
        return list(self.g.predecessors(node_id))

    def children(self, node_id):
        """
        Gets the children for the specified node ID.

        :param node_id: Node ID.
        :return: List of children node IDs.
        """
        return list(self.g.successors(node_id))

    def coparents(self, node_id):
        """
        Gets the coparents of the specified node.

        :param node_id:  Node ID.
        :return: List of coparent node IDS.
        """
        copas = set()
        for ch in self.children(node_id):
            pas = self.parents(ch)
            for pa in pas:
                copas.add(pa)
        return list(copas)

    def markov_blanket(self, node_id):
        """
        Gets the Markov blanket of the specified node. The Markov blanket of a node is defined as its parents,
        children, and co-parents.

        :param node_id: Node ID.
        :return: List of node IDs in the Markov blanket.
        """
        blanket = set()

        for pa in self.parents(node_id):
            blanket.add(pa)
        for ch in self.children(node_id):
            blanket.add(ch)
        for copa in self.coparents(node_id):
            blanket.add(copa)

        if node_id in blanket:
            blanket.remove(node_id)
        return sorted(list(blanket))

    def get_sorted_topology(self):
        """
        Gets the sorted topology of the DAG.

        :return: Sorted topology of the DAG.
        """
        return list(topological_sort(self.g))


class Parameters(object):
    """
    Parameters, which are made up of the means and covariance matrix.
    """

    def __init__(self, means, cov):
        """
        Ctor.

        :param means: Means.
        :param cov: Covariance matrix.
        """
        self.means = means
        self.cov = cov

        if len(means) != cov.shape[0]:
            raise ValueError('number of means {} != number of variables {}'.format(len(means), cov.shape[0]))
        if cov.shape[0] != cov.shape[1]:
            raise ValueError('covariance matrix is not square: {} x {}'.format(cov.shape[0], cov.shape[1]))

    def get_mean(self, node_id):
        """
        Gets the mean associated with the node ID.

        :param node_id: Integer node ID.
        :return: Mean.
        """
        return self.means[node_id]

    def get_stdev(self, node_id):
        """
        Gets the standard deviation associated with the node ID.

        :param node_id: Integer node ID.
        :return: Variance.
        """
        return np.sqrt(self.cov[node_id, node_id])


class Bbn(object):
    """
    Bayesian Belief Network.
    """

    def __init__(self, dag, params):
        """
        Ctor. Note that the dimensions of the DAG and parameters must match.

        :param dag: DAG.
        :param params: Parameters (means and covariance).
        """
        self.dag = dag
        self.params = params
        self.mvn = MvnInference(params.means, params.cov)
        self.evidence = [None for _ in range(dag.number_of_nodes())]

        # all of these checks are to make sure the DAG and params align
        # check for 1-to-1 correspondence between nodes and variables
        num_nodes = dag.number_of_nodes()
        num_means = len(params.means)
        dag_nodes = dag.nodes()
        min_node_id = min(dag_nodes)
        max_node_id = max(dag_nodes)

        # num_vars1 = params.cov.shape[0]
        # num_vars2 = params.cov.shape[1]

        if num_nodes != num_means:
            raise ValueError('number of nodes {} != number of means {}'.format(num_nodes, num_means))
        # if num_nodes != num_vars1:
        #     raise ValueError('number of nodes {} != number of variables {}'.format(num_nodes, num_vars1))
        # if num_vars1 != num_vars2:
        #     raise ValueError('covariance is not a square; rows = {}, cols = {}'.format(num_vars1, num_vars2))
        if 0 != min_node_id:
            raise ValueError('Node IDs should start with zero, but it started with {}'.format(min_node_id))
        if num_nodes - 1 != max_node_id:
            raise ValueError('Node IDs should end with number_of_nodes - 1, but it was {}'.format(max_node_id))

        # make sure the node IDs are sequential
        for e, o in zip(range(num_nodes), sorted(dag_nodes)):
            if e != o:
                raise ValueError('Node IDs should be sequential integers but was not: {}'.format(sorted(dag.nodes())))

    def set_evidence(self, node_id, e):
        """
        Sets the evidence for the specified node.

        :param node_id: Node ID.
        :param e: Evidence.
        :return: BBN.
        """
        self.evidence[node_id] = e
        return self

    def get_evidence(self, node_id):
        """
        Gets the evidence for the specified node.

        :param node_id: Node ID.
        :return: Evidence.
        """
        return self.evidence[node_id]

    def get_evidences(self):
        """
        Gets all the evidences.

        :return: Array of evidences.
        """
        return self.evidence

    def clear_evidences(self):
        """
        Clears all the evidences.

        :return: None
        """
        self.evidence = [None for _ in range(self.dag.number_of_nodes())]

    def __has_parents__(self, node_id):
        """
        Checks if the specified node has parents.

        :param node_id: Node ID.
        :return: A boolean indicating if the node has parents.
        """
        return len(self.dag.parents(node_id)) > 0

    def __has_evidence__(self, node_id):
        """
        Checks if the specified node has evidence.

        :param node_id: Node ID.
        :return: A boolean indicating if the node has evidence.
        """
        return False if np.isnan(self.evidence[node_id]) else True

    def __get_evidences__(self):
        evidences = [(i, v) for i, v in enumerate(self.evidence)]
        evidences = list(filter(lambda tup: tup[1] is not None, evidences))
        v = list(map(lambda tup: tup[1], evidences))
        iv = list(map(lambda tup: tup[0], evidences))
        return v, iv

    def do_inference(self, N=None):
        """
        Conducts inference.

        :param N: Number of samples. If an integer is provided, the sampled means and covariances will be returned.
        :return: (means, covariances)
        """
        v, iv = self.__get_evidences__()
        self.mvn.update_mean_cov(v, iv)

        M, S = self.mvn.get_params()

        if N is not None:
            D = multivariate_normal.rvs(M, S, N)
            M = D.mean(axis=0)
            S = np.cov(D.T)

        return M, S

    def get_memmap(self, X):
        import os
        from joblib import dump, load

        folder = './joblib_memmap'
        try:
            os.mkdir(folder)
        except FileExistsError:
            pass

        data_filename_memmap = os.path.join(folder, 'data_memmap')
        dump(X, data_filename_memmap)
        data = load(data_filename_memmap, mmap_mode='r')
        return data

    def predict_proba(self, X, n_jobs=-1, batch_size=5000, parallel=True):
        """
        Predict probability.

        :param X: data.
        :param n_jobs: Number of jobs.
        :param batch_size: Batch size.
        :param parallel: Parallel computation?
        :return: Probabilities.
        """
        num_data = X.shape[0]
        num_nodes = self.dag.number_of_nodes()
        D = self.get_memmap(X)

        all_units = []
        for row_id in range(num_data):
            units = []
            for node_id in range(num_nodes):
                if self.__has_parents__(node_id) is False:
                    m = self.params.means[node_id]
                    s = math.sqrt(self.params.cov[node_id, node_id])
                    unit = WORK_UNIT(row_id, node_id, m, s, None)
                    units.append(unit)
                else:
                    m = self.params.means
                    s = self.params.cov
                    dep = sorted(self.dag.parents(node_id))
                    unit = WORK_UNIT(row_id, node_id, m, s, dep)
                    units.append(unit)
            all_units.append(WORK_UNITS(row_id, units))

        jobs = n_jobs if parallel is True else 1
        results = Parallel(n_jobs=jobs,
                           backend='threading',
                           batch_size=batch_size)(delayed(__get_probs__)(D, W) for W in all_units)
        results = sorted(results, key=lambda r: r.row_id)
        results = list(map(lambda r: r.prob, results))

        return np.array(results)

    def predict_log_proba(self, X, n_jobs=-1, batch_size=5000, parallel=True):
        """
        Predict log probabilities.

        :param X: Data.
        :param n_jobs: Number of jobs.
        :param batch_size: Batch size.
        :param parallel: Compute in parallel?
        :return: Log probabilities.
        """
        num_data = X.shape[0]
        num_nodes = self.dag.number_of_nodes()
        D = self.get_memmap(X)

        all_units = []
        for row_id in range(num_data):
            units = []
            for node_id in range(num_nodes):
                if self.__has_parents__(node_id) is False:
                    m = self.params.means[node_id]
                    s = math.sqrt(self.params.cov[node_id, node_id])
                    unit = WORK_UNIT(row_id, node_id, m, s, None)
                    units.append(unit)
                else:
                    m = self.params.means
                    s = self.params.cov
                    dep = sorted(self.dag.parents(node_id))
                    unit = WORK_UNIT(row_id, node_id, m, s, dep)
                    units.append(unit)
            all_units.append(WORK_UNITS(row_id, units))

        jobs = n_jobs if parallel is True else 1
        results = Parallel(n_jobs=jobs,
                           backend='threading',
                           batch_size=batch_size)(delayed(__get_log_probs__)(D, W) for W in all_units)
        results = sorted(results, key=lambda r: r.row_id)
        results = list(map(lambda r: r.prob, results))

        return np.array(results)

    def log_prob(self, X):
        """
        Computes the log probabilities.

        :param X: Data.
        :return: Log probabilities.
        """
        num_nodes = self.dag.number_of_nodes()
        sum = 0.0
        for node_id in range(num_nodes):
            if self.__has_parents__(node_id) is False:
                d = X[:, node_id]
                m = self.params.means[node_id]
                s = np.sqrt(self.params.cov[node_id, node_id])
                logp = np.sum([np.log10(p) for p in dnorm(d, m, s)])
                sum += logp
            else:
                m = self.params.means
                s = self.params.cov
                dep = sorted(self.dag.parents(node_id))
                logp = np.sum([np.log10(p) for p in dcmvnorm(X, m, s, node_id, dep)])
                sum += logp
        return sum



from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.graph import Ug


class Moralizer(object):
    """
    Graph moralizer for a DAG.
    """

    @staticmethod
    def moralize(dag):
        """
        Moralizes a DAG.

        :param dag: DAG.
        :return: Moralized (undirected) graph.
        """
        ug = Ug()
        for node in dag.get_nodes():
            ug.add_node(node)
        for edge in dag.get_edges():
            ug.add_edge(Edge(edge.i, edge.j, EdgeType.UNDIRECTED))
        for node in dag.get_nodes():
            parents = [dag.get_node(pa) for pa in dag.get_parents(node.id)]
            size = len(parents)
            for i in range(size):
                pa1 = parents[i]
                for j in range(i + 1, size):
                    pa2 = parents[j]
                    ug.add_edge(Edge(pa1, pa2, EdgeType.UNDIRECTED))
        for node in dag.get_nodes():
            parents = [dag.get_node(pa) for pa in dag.get_parents(node.id)]
            node.add_metadata('parents', parents)

        return ug



from pybbn.graph.potential import PotentialUtil


class EvidenceDistributor(object):
    """
    Evidence distributor. Passes messages using breadth-first-search (BFS).
    Messages are passed from the start clique to the far remote cliques.
    """

    def __init__(self, join_tree, start_clique):
        """
        Ctor.

        :param join_tree: Join tree.
        :param start_clique: Start clique.
        """
        self.join_tree = join_tree
        self.start_clique = start_clique

    @staticmethod
    def __get_neighboring_cliques__(join_tree, clique):
        """
        Gets the neighboring cliques.
        :param join_tree: Join tree.
        :param clique: Clique.
        :return: Tuple. First tuple is list of neighboring sep-sets and second is list of neighboring cliques.
        """
        sepsets = []
        cliques = []

        sep_set_ids = join_tree.get_neighbors(clique.id)
        for sep_set_id in sep_set_ids:
            clique_ids = join_tree.get_neighbors(sep_set_id)
            for clique_id in clique_ids:
                t1 = (sep_set_id, join_tree.get_node(sep_set_id))
                t2 = (clique_id, join_tree.get_node(clique_id))
                if clique_id != clique.id:
                    sepsets.append(t1)
                    cliques.append(t2)

        return sepsets, cliques

    def start(self):
        """
        Starts the evidence distribution.
        """
        self.start_clique.mark()
        sepsets, cliques = self.__get_neighboring_cliques__(self.join_tree, self.start_clique)

        for clique in cliques:
            clique[1].mark()

        for s, c in zip(sepsets, cliques):
            PotentialUtil.pass_single_message(self.join_tree, self.start_clique, s[1], c[1])
            self.__walk__(self.start_clique, s[1], c[1])

    def __walk__(self, x, s, y):
        """
        Walks away from the specified node y.

        :param x: Clique.
        :param s: Separation-set.
        :param y: Clique.
        """
        sepsets, cliques = self.__get_neighboring_cliques__(self.join_tree, y)

        s_arr = []
        c_arr = []
        for sep, cli in zip(sepsets, cliques):
            if not cli[1].is_marked():
                cli[1].mark()
                s_arr.append(sep)
                c_arr.append(cli)

        for sep, cli in zip(s_arr, c_arr):
            PotentialUtil.pass_single_message(self.join_tree, y, sep[1], cli[1])
            self.__walk__(y, sep[1], cli[1])



import math

import numpy as np
from numpy.linalg import inv, det
from scipy import linalg


def __get_box_muller_sample__():
    """
    Gets a sample using the Box-Muller transform.

    :return: A sample point.
    """
    r = 0
    x = 0
    y = 0

    while True:
        x = 2.0 * np.random.uniform(0.0, 1.0, 1)[0] - 1.0
        y = 2.0 * np.random.uniform(0.0, 1.0, 1)[0] - 1.0
        r = (x * x) + (y * y)

        if 0.0 < r <= 1.0:
            break

    z = x * math.sqrt(-2.0 * math.log(r) / r)
    return z


def __get_box_muller_samples__(n):
    """
    Gets n independent Box-Muller samples.

    :param n: Number of samples.
    :return: Array of Box-Muller samples of dimension 1 x n.
    """
    return np.array([__get_box_muller_sample__() for _ in range(n)])


def __get_sample__(m, s):
    """
    Sample from the Gaussian distribution with mean=m and standard
    deviation=s.

    :param m: Mean.
    :param s: Standard deviation.
    :return: Sample point.
    """
    z = __get_box_muller_sample__()
    return z * s + m


def __slice_acov__(cov, dep, given):
    """
    Slices a covariance matrix keeping only the row associated with the dependent variable
    minus its self-covariance.

    :param cov: Covariance matrix.
    :param dep: Index of dependent variable.
    :param given: Array of indices of independent variables.
    :return: A 1 x |given| vector of covariances.
    """
    row_selector = dep
    col_selector = [x for x in range(cov.shape[1]) if x in given]
    v = cov[row_selector, col_selector]
    return v


def __slice_scov__(cov, dep, given):
    """
    Slices a covariance matrix keeping only the covariances between the variables
    indicated by the array of indices of the independent variables.

    :param cov: Covariance matrix.
    :param dep: Index of dependent variable.
    :param given: Array of indices of independent variables.
    :return: A |given| x |given| matrix of covariance.
    """
    row_selector = [x for x in range(cov.shape[0]) if x in given and x != dep]
    col_selector = [x for x in range(cov.shape[1]) if x in given and x != dep]
    v = cov[row_selector, :]
    v = v[:, col_selector]
    return v


def rnorm(n, m, s):
    """
    Sample from the Gaussian distribution with mean=m and standard
    deviation=s.

    :param n: Number of samples.
    :param m: Mean.
    :param s: Standard deviation.
    :return: Sample points.
    """
    for i in range(n):
        yield __get_sample__(m, s)


def dnorm(data, m, s):
    """
    Gets the probability of each value in x given the mean=m and standard deviation=s.

    :param data: Array of values.
    :param m: Mean.
    :param s: Standard deviation.
    :return: Probabilities.
    """
    c = 1.0 / math.sqrt(2.0 * math.pi * math.pow(s, 2.0))
    d = 2.0 * math.pow(s, 2.0)
    for x in data:
        exponent = -1.0 * math.pow(x - m, 2.0) / d
        yield c * math.exp(exponent)


def rcmvnorm(n, m, cov, dep, given, X):
    """
    Samples from the conditional multivariate Gaussian distribution with means=m and
    covariance matrix=cov subject to the values X.

    :param n: The number of samples to generate.
    :param m: An array of means.
    :param cov: Covariance matrix.
    :param dep: Index of dependent variable.
    :param given: Array of indices of independent variables.
    :param X: Values of dependent variables.
    :return: Sample points.
    """
    # cov.rows should equal cov.cols
    # |m| should equal cov.rows
    # dep should be one integer
    # |given| should equal X.cols
    # |given| < cov.cols

    col_selector = [x for x in range(cov.shape[1]) if x in given]

    cov_yy = cov[dep, dep]
    cov_yx = __slice_acov__(cov, dep, given)
    cov_xx = inv(__slice_scov__(cov, dep, given))

    m_x = m[col_selector]
    m_y = m[dep]

    cov_yx_dot_cov_xx = cov_yx.dot(cov_xx)
    v_y = cov_yy - cov_yx.dot(cov_xx).dot(cov_yx.transpose())

    for i in range(n):
        e_y = m_y + cov_yx_dot_cov_xx.dot(X - m_x)
        y = list(rnorm(1, e_y, v_y))[0]
        yield y


def dcmvnorm(data, m, cov, dep, given):
    """
    Gets the probability of the dependent variable given the independent ones.

    :param data: Matrix of data.
    :param m: Array of means.
    :param cov: Covariance matrix.
    :param dep: Index of dependent variable.
    :param given: Array of indices of independent variables.
    :return: Probabilities.
    """
    # cov.rows should equal cov.cols
    # |m| should equal cov.rows
    # dep should be one integer
    # |given| should equal X.cols
    # |given| < cov.cols

    col_selector = [x for x in range(cov.shape[1]) if x in given]
    X = data[:, col_selector]
    y = data[:, dep]

    cov_yy = cov[dep, dep]
    cov_yx = __slice_acov__(cov, dep, given)
    cov_xx = inv(__slice_scov__(cov, dep, given))

    m_x = m[col_selector]
    m_y = m[dep]

    cov_yx_dot_cov_xx = cov_yx.dot(cov_xx)
    v_y = cov_yy - cov_yx.dot(cov_xx).dot(cov_yx.transpose())

    for i in range(data.shape[0]):
        e_y = m_y + cov_yx_dot_cov_xx.dot(X[i] - m_x)
        d = y[i]
        p_y = list(dnorm([d], e_y, v_y))[0]
        yield p_y


def rmvnorm(n, m, cov):
    """
    Samples from the multivariate Gaussian distribution with means=m and covariance matrix=cov.

    :param n: Number of samples.
    :param m: Array of means.
    :param cov: Covariance matrix.
    :return: Sample points.
    """
    A = linalg.cholesky(cov).transpose()
    v = cov.shape[0]
    for i in range(n):
        yield m + A.dot(__get_box_muller_samples__(v))


def dmvnorm(data, m, cov):
    """
    Computes the probabilities of the sample points in X.

    :param data: Data matrix.
    :param m: Means.
    :param cov: Covariance matrix.
    :return: Probabilities.
    """
    cov_inv = inv(cov)
    cov_det = det(cov)
    k = cov.shape[0]
    d = math.sqrt(math.pow(2.0 * math.pi, k) * cov_det)
    for i in range(data.shape[0]):
        x = data[i] - m
        y = x.reshape((k, 1))
        e = -0.5 * x.dot(cov_inv).dot(y)
        yield math.exp(e) / d


class RandCondMvn(object):
    """
    Random conditional multivariate normal.
    """

    def __init__(self, m, cov, dep, given):
        """
        Ctor.

        :param m: An array of means.
        :param cov: Covariance matrix.
        :param dep: Index of dependent variable.
        :param given: Array of indices of independent variables.
        :return: None.
        """
        # cov.rows should equal cov.cols
        # |m| should equal cov.rows
        # dep should be one integer
        # |given| should equal X.cols
        # |given| < cov.cols

        col_selector = [x for x in range(cov.shape[1]) if x in given]

        cov_yy = cov[dep, dep]
        cov_yx = __slice_acov__(cov, dep, given)
        cov_xx = inv(__slice_scov__(cov, dep, given))

        self.m_x = m[col_selector]
        self.m_y = m[dep]

        self.cov_yx_dot_cov_xx = cov_yx.dot(cov_xx)
        self.v_y = cov_yy - cov_yx.dot(cov_xx).dot(cov_yx.transpose())

    def next(self, X):
        """
        Samples from the conditional multivariate Gaussian distribution

        :param X: Values of dependent variables.
        :return: Sample.
        """
        e_y = self.m_y + self.cov_yx_dot_cov_xx.dot(X - self.m_x)
        y = list(rnorm(1, e_y, self.v_y))[0]
        return y



from functools import reduce
from itertools import combinations

from pybbn.graph.edge import Edge, EdgeType
from pybbn.graph.graph import Ug
from pybbn.graph.node import Clique


class Triangulator(object):
    """
    Triangulator. Triangulates an undirected moralized graph and produces cliques in the process.
    """

    @staticmethod
    def triangulate(m):
        """
        Triangulates the specified moralized graph.

        :param m: Moralized undirected graph.
        :return: Array of cliques.
        """
        cliques = []
        mm = Triangulator.duplicate(m)
        while len(mm.get_nodes()) > 0:
            node_clique = Triangulator.select_node(mm)
            clique = Clique(node_clique.get_bbn_nodes())

            if not Triangulator.is_subset(cliques, clique):
                cliques.append(clique)

            mm.remove_node(node_clique.node.id)

            for edge in node_clique.edges:
                m.add_edge(edge)
                mm.add_edge(edge)

        return cliques

    @staticmethod
    def duplicate(g):
        """
        Duplicates a undirected graph.

        :param g: Undirected graph.
        :return: Undirected graph.
        """
        ug = Ug()
        for node in g.get_nodes():
            ug.add_node(node)
        for edge in g.get_edges():
            ug.add_edge(edge)
        return ug

    @staticmethod
    def generate_cliques(m):
        """
        Generates a list of node cliques.

        :param m: Graph.
        :return: List of NodeCliques.
        """

        def get_neighbors(node, m):
            return [m.get_node(neighbor_id) for neighbor_id in m.get_neighbors(node.id)]

        def get_weight(node, m):
            return Triangulator.get_weight(node, m)

        def get_edges_to_add(node, m):
            return Triangulator.get_edges_to_add(node, m)

        return (NodeClique(node, get_neighbors(node, m), get_weight(node, m), get_edges_to_add(node, m))
                for node in m.get_nodes())

    @staticmethod
    def select_node(m):
        """
        Selects a clique from the specified graph. Cliques are sorted by number of edges, weight, and id (asc).

        :param m: Graph.
        :return: Clique.
        """
        return sorted(Triangulator.generate_cliques(m), key=lambda x: (len(x.edges), x.weight, x.node.id))[0]

    @staticmethod
    def get_weight(n, m):
        """
        Gets the weight of a BBN node. The weight of a node is the product of the its weight with all its
        neighbors' weight.

        :param n: BBN node.
        :param m: Graph.
        :return: Weight.
        """
        if len(m.neighbors[n.id]) == 0:
            return n.get_weight()
        weights = (m.get_node(neighbor_id).get_weight() for neighbor_id in m.get_neighbors(n.id))
        return n.get_weight() * reduce(lambda x, y: x * y, weights)

    @staticmethod
    def get_edges_to_add(n, m):
        """
        Gets edges to add.

        :param n: BBN node.
        :param m: Graph.
        :return: Array of edges.
        """
        neighbors = [m.get_node(i) for i in m.get_neighbors(n.id)]
        return [Edge(neighbors[i], neighbors[j], EdgeType.UNDIRECTED)
                for i, j in combinations(range(len(neighbors)), 2)
                if not m.edge_exists(neighbors[i].id, neighbors[j].id)]

    @staticmethod
    def is_subset(cliques, clique):
        """
        Checks if the specified clique is a subset of the specified list of cliques.

        :param cliques: List of cliques.
        :param clique: Clique.
        :return: A boolean indicating if the clique is a subset.
        """
        for i in range(len(cliques)):
            if cliques[i].is_superset(clique):
                return True
        return False


class NodeClique:
    """
    Node clique.
    """

    def __init__(self, node, neighbors, weight, edges):
        """
        Ctor.

        :param node: BBN node.
        :param neighbors: BBN nodes (neighbors).
        :param weight: Weight.
        :param edges: Edges.
        """
        self.node = node
        self.neighbors = neighbors
        self.weight = weight
        self.edges = edges

    def get_bbn_nodes(self):
        """
        Gets all the BBN nodes in this node clique.

        :return: Array of BBN nodes.
        """
        neighbors = [node for node in self.neighbors] + [self.node]
        return neighbors

    def __str__(self):
        return f'{self.node.id}|weight={self.weight}|edges={len(self.edges)}|neighbors={len(self.neighbors)}'



import warnings
from collections import namedtuple

import numpy as np
from numpy.linalg import inv
from numpy.random import multivariate_normal, normal
from scipy.stats import multivariate_normal as mvn_normal

COV = namedtuple('COV', 'C11 C12 C21 C22 C22I')


def __compute_means__(v, M, C, i1, i2):
    """
    Computes the conditional means.

    :param v: The values conditioned on.
    :param M: Marginal means (vector).
    :param C: Covariance matrix.
    :param i1: Indices of unconditioned variables.
    :param i2: Indices of conditioned variables.
    :return: Conditional means.
    """
    return M[i1] + C.C12.dot(C.C22I).dot(v - M[i2])


def __compute_covs__(C):
    """
    Computes the conditional covariance matrix.

    :param C: Partitioned covariance matrix.
    :return: Conditional covariance matrix.
    """
    return C.C11 - C.C12.dot(C.C22I).dot(C.C21)


def __update_mean__(m, v, M, i1, i2):
    """
    Updates the means.

    :param m: Means of unconditioned variables.
    :param v: Means of conditioned variables.
    :param M: Vector of marginal means.
    :param i1: Indices of unconditioned variables.
    :param i2: Indices of conditioned variables.
    :return: Updated means.
    """
    M_u = np.copy(M)
    for i, mu in zip(i1, m):
        M_u[i] = mu
    for i, mu in zip(i2, v):
        M_u[i] = mu
    return M_u


def __update_cov__(c, S, i1, i2):
    """
    Updates the covariance matrix.

    :param c: Conditioned covariance matrix.
    :param S: Covariance matrix.
    :param i1: Indices of unconditioned variables.
    :param i2: Indices of conditioned variables.
    :return: Updated covariance matrix.
    """
    m = np.copy(S)
    rows, cols = c.shape
    for row in range(rows):
        for col in range(cols):
            m[i1[row], i1[col]] = c[row, col]
    for i in i2:
        m[i, i] = 0.01
    return m


def __to_row_indices__(indices):
    """
    Creates row indices for help in broadcasting/slicing matrix.

    :param indices: Indices.
    :return: Array of array of indices.
    """
    return [[i] for i in indices]


def __to_col_indices__(indices):
    """
    Creates column indices for help in broadcasting/slicing matrix.

    :param indices: Indices.
    :return: Array.
    """
    return indices


def get_covariances(i1, i2, S):
    """
    Gets the partitioned matrices of S.

    :param i1: Indices of unconditioned variables.
    :param i2: Indices of conditioned variables.
    :param S: Covariance matrix.
    :return: Partitioned matrices of S.
    """
    r = __to_row_indices__(i1)
    c = __to_col_indices__(i1)
    C11 = S[r, c]

    r = __to_row_indices__(i1)
    c = __to_col_indices__(i2)
    C12 = S[r, c]

    r = __to_row_indices__(i2)
    c = __to_col_indices__(i1)
    C21 = S[r, c]

    r = __to_row_indices__(i2)
    c = __to_col_indices__(i2)
    C22 = S[r, c]

    C22I = inv(C22)

    return COV(C11, C12, C21, C22, C22I)


class MvnInference(object):
    """
    Multivariate normal Gaussian.
    """

    def __init__(self, M, S, N=10000):
        """
        Ctor.

        :param M: Vector of means.
        :param S: Matrix of covariances.
        :param N: Number of samples (default: 10,000).
        """
        self.M = M
        self.S = S
        self.N = N
        self.M_u = None
        self.S_u = None

    def clear(self):
        """
        Clears the updated mean and covariance.

        :return: None.
        """
        self.M_u = None
        self.S_u = None

    def update_mean_cov(self, v, iv):
        """
        Updates the means and covariances.

        :param v: Values conditioned.
        :param iv: Indices of variables conditioned on.
        :return: None.
        """
        if v is None or iv is None or len(v) == 0 or len(iv) == 0:
            self.M_u = np.copy(self.M)
            self.S_u = np.copy(self.S)
            return

        if self.M.shape[0] == 1:
            self.M_u = np.array([v[0]])
            self.S_u = 0.01
            return

        i2 = iv.copy()
        i1 = [i for i in range(self.S.shape[0]) if i not in i2]

        C = get_covariances(i1, i2, self.S)
        m = __compute_means__(v, self.M, C, i1, i2)
        c = __compute_covs__(C)
        self.M_u = __update_mean__(m, v, self.M, i1, i2)
        self.S_u = __update_cov__(c, self.S, i1, i2)

    def get_params(self):
        """
        Gets the means and covariances.

        :return: (means, covariances).
        """
        return (self.M, self.S) if self.M_u is None or self.S_u is None else (self.M_u, self.S_u)

    def get_samples(self):
        """
        Gets drawn samples from the current means and covariances.

        :return:
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            M, S = self.get_params()

            if M.shape[0] == 1:
                return normal(M[0], S, self.N)

            return multivariate_normal(M, S, self.N)

    def get_corr(self):
        """
        Estimates the pair-wise correlation between all variables
        using drawn samples from the means and covariances.

        :return: Pairwise correlation matrix.
        """
        return np.corrcoef(self.get_samples().T)

    def predict_proba(self, X):
        """
        Predicts the probabilities.

        :param X: Data.
        :return: Probabilities.
        """
        M, S = self.get_params()
        return multivariate_normal.pdf(X, mean=M, cov=S)

    def predict_log_proba(self, X):
        """
        Predicts the log probabilities.

        :param X: Data.
        :return: Log probabilities.
        """
        M, S = self.get_params()
        return mvn_normal.logpdf(X, mean=M, cov=S)



from collections import defaultdict
from copy import deepcopy
from enum import Enum

from pybbn.graph.edge import JtEdge
from pybbn.graph.graph import Ug
from pybbn.graph.node import SepSet, Clique, BbnNode
from pybbn.graph.potential import Potential, PotentialEntry, PotentialUtil
from pybbn.graph.variable import Variable


class JoinTree(Ug):
    """
    Join tree.
    """

    def __init__(self):
        """
        Ctor.
        """
        Ug.__init__(self)
        self.potentials = dict()
        self.evidences = dict()
        self.listener = None
        self.parent_info = defaultdict(set)
        # self.__all_nodes__ = None

    def __deepcopy__(self, memodict={}):
        nodes = deepcopy(self.nodes, memodict)
        edges = deepcopy(self.edges, memodict)
        edge_map = deepcopy(self.edge_map, memodict)
        neighbors = deepcopy(self.neighbors, memodict)
        potentials = deepcopy(self.potentials, memodict)
        evidences = deepcopy(self.evidences, memodict)
        parent_info = deepcopy(self.parent_info, memodict)

        jt = JoinTree()
        jt.nodes = nodes
        jt.edges = edges
        jt.edge_map = edge_map
        jt.neighbors = neighbors
        jt.potentials = potentials
        jt.evidences = evidences
        jt.parent_info = parent_info
        return jt

    def get_posteriors(self):
        """
        Gets the posterior for all nodes.

        :return: Map. Keys are node names; values are map of node values to posterior probabilities.
        """
        bbn_nodes = self.get_bbn_nodes()

        posteriors = {}

        for bbn_node in bbn_nodes:
            potential = self.get_bbn_potential(bbn_node)

            m = {}
            for potential_entry in potential.entries:
                k = ''.join([f'{y}' for _, y in potential_entry.entries.items()])
                m[k] = potential_entry.value

            name = bbn_node.variable.name
            posteriors[name] = m

        return posteriors

    def get_bbn_potential(self, node):
        """
        Gets the potential associated with the specified BBN node.

        :param node: BBN node.
        :return: Potential.
        """
        clique = node.metadata['parent.clique']
        potential = PotentialUtil.normalize(PotentialUtil.marginalize_for(self, clique, [node]))
        return potential

    def unmark_cliques(self):
        """
        Unmarks the cliques.
        """
        for clique in self.get_cliques():
            clique.unmark()

    def update_bbn_cpts(self, cpts):
        """
        Updates the CPTs of the BBN nodes.

        :param cpts: Dictionary of CPTs. Keys are ids of BBN node and values are new CPTs.
        :return: None
        """
        bbn_nodes = {node.id: node for clique in self.get_cliques() for node in clique.nodes}
        for idx, cpt in cpts.items():
            if idx in bbn_nodes:
                bbn_nodes[idx].probs = cpt
                bbn_nodes[idx].potential = None

    def get_bbn_node_and_parents(self):
        """
        Gets a map of nodes and its parents.

        :return: Map. Keys are node ID and values are list of nodes.
        """
        bbn_nodes = {node.id: node for clique in self.get_cliques() for node in clique.nodes}
        result = {node: [pa for pa_id, pa in bbn_nodes.items() if pa_id in self.parent_info[node_id]]
                  for node_id, node in bbn_nodes.items()}
        return result

    def __get_bbn_nodes__(self):
        """
        Gets all BBN nodes (cached).

        :return: Dictionary of BBN nodes.
        """
        # if self.__all_nodes__ is None:
        #     self.__all_nodes__ = {node.id: node for clique in self.get_cliques() for node in clique.nodes}
        # return self.__all_nodes__
        result = {node.id: node for clique in self.get_cliques() for node in clique.nodes}
        return result

    def get_bbn_nodes(self):
        """
        Gets all the BBN nodes in this junction tree.

        :return: List of BBN nodes.
        """
        return list(self.__get_bbn_nodes__().values())

    def get_bbn_node(self, id):
        """
        Gets the BBN node associated with the specified id.

        :param id: Node id.
        :return: BBN node or None if no such node exists.
        """
        bbn_nodes = self.__get_bbn_nodes__()
        if id in bbn_nodes:
            return bbn_nodes[id]
        return None

    def get_bbn_node_by_name(self, name):
        """
        Gets the BBN node associated with the specified name.

        :param name: Node name.
        :return: BBN node or None if no such node exists.
        """
        bbn_nodes = {node.variable.name: node for clique in self.get_cliques() for node in clique.nodes}
        if name in bbn_nodes:
            return bbn_nodes[name]
        return None

    def find_cliques_with_node_and_parents(self, id):
        """
        Finds all cliques in this junction tree having the specified node and its parents.

        :param id: Node id.
        :return: Array of cliques.
        """
        ids = self.__get_parent_ids__(id)
        ids.append(id)

        set1 = set(ids)
        result = [clique for clique in self.get_cliques() if clique.get_node_ids().issuperset(set1)]
        return result

    def add_potential(self, clique, potential):
        """
        Adds a potential associated with the specified clique.

        :param clique: Clique.
        :param potential: Potential.
        :return: This join tree.
        """
        self.potentials[clique.id] = potential
        return self

    def get_cliques(self):
        """
        Gets all the cliques in this junction tree.

        :return: Array of cliques.
        """
        return [clique for clique in self.get_nodes() if not isinstance(clique, SepSet)]

    def get_sep_sets(self):
        """
        Gets all the separation sets in this junction tree.

        :return: Array of separation sets.
        """
        return [sep_set for sep_set in self.get_nodes() if isinstance(sep_set, SepSet)]

    def add_edge(self, edge):
        """
        Adds an JtEdge.

        :param edge: JtEdge.
        :return: This join tree.
        """
        if not isinstance(edge, JtEdge):
            return self

        sep_set = edge.sep_set
        lhs = edge.i
        rhs = edge.j

        if self.__shouldadd__(edge):
            self.add_node(sep_set)
            self.add_node(lhs)
            self.add_node(rhs)

            self.edge_map[lhs.id].add(sep_set.id)
            self.edge_map[rhs.id].add(sep_set.id)
            self.neighbors[lhs.id].add(sep_set.id)
            self.neighbors[rhs.id].add(sep_set.id)

            self.edge_map[sep_set.id].add(lhs.id)
            self.edge_map[sep_set.id].add(rhs.id)
            self.neighbors[sep_set.id].add(lhs.id)
            self.neighbors[sep_set.id].add(rhs.id)

            self.edges[edge.key] = edge

        return self

    def get_flattened_edges(self):
        """
        Gets all the edges "flattened" out. Since separation-sets are really hyper-edges, this method breaks
        separation-sets into two edges.

        :return: Array of edges.
        """
        edges = []
        for edge in self.edges.values():
            edges.append(edge.get_lhs_edge())
            edges.append(edge.get_rhs_edge())
        return edges

    def set_listener(self, listener):
        """
        Sets the listener.

        :param listener: JoinTreeListener.
        """
        self.listener = listener

    def get_evidence(self, node, value):
        """
        Gets the evidence associated with the specified BBN node and value.

        :param node: BBN node.
        :param value: Value.
        :return: Potential (the evidence).
        """
        if node.id not in self.evidences:
            self.evidences[node.id] = dict()

        if value not in self.evidences[node.id]:
            entry = PotentialEntry()
            entry.add(node.id, value)
            entry.value = 1.0

            potential = Potential()
            potential.add_entry(entry)

            self.evidences[node.id][value] = potential

        result = self.evidences[node.id][value]
        return result

    def get_change_type(self, evidences):
        """
        Gets the change type associated with the specified list of evidences.

        :param evidences: List of evidences.
        :return: ChangeType.
        """
        changes = []
        for evidence in evidences:
            node = evidence.node
            potentials = self.evidences[node.id]
            change = evidence.compare(potentials)
            changes.append(change)

        count = len([change_type for change_type in changes if ChangeType.RETRACTION == change_type])
        if count > 0:
            return ChangeType.RETRACTION

        count = len([change_type for change_type in changes if ChangeType.UPDATE == change_type])
        if count > 0:
            return ChangeType.UPDATE

        return ChangeType.NONE

    def get_unobserved_evidence(self, node):
        """
        Gets the unobserved evidences associated with the specified node.

        :param node: BBN node.
        :return: Evidence.
        """
        evidence = Evidence(node, EvidenceType.UNOBSERVE)
        for value in node.variable.values:
            evidence.add_value(value, 1.0)
        return evidence

    def unobserve(self, nodes):
        """
        Unobserves a list of nodes.

        :param nodes: List of nodes.
        :return: This join tree.
        """
        evidences = [self.get_unobserved_evidence(node) for node in nodes]
        self.update_evidences(evidences)
        return self

    def unobserve_all(self):
        """
        Unobserves all BBN nodes.

        :return: This join tree.
        """
        self.unobserve(self.get_bbn_nodes())
        return self

    def update_evidences(self, evidences):
        """
        Updates this join tree with the list of specified evidence.

        :param evidences: List of evidences.
        :return: This join tree.
        """
        for evidence in evidences:
            evidence.validate()
        change = self.get_change_type(evidences)
        for evidence in evidences:
            node = evidence.node
            potentials = self.evidences[node.id]

            for k, v in evidence.values.items():
                potential = potentials[k]
                potential.entries[0].value = v
        self.__notify_listener__(change)
        return self

    def set_observation(self, evidence):
        """
        Sets a single observation.

        :param evidence: Evidence.
        :return: This join tree.
        """
        potentials = self.evidences[evidence.node.id]

        pvalues = []
        for v, potential in potentials.items():
            entry = potential.entries[0]
            p = entry.value
            if 1.0 == p:
                pvalues.append(v)

        cvalues = []
        for v, likelihood in evidence.values.items():
            if 1.0 == likelihood:
                cvalues.append(v)

        if 1 == len(pvalues):
            last_value = pvalues[0]
            curr_value = cvalues[0]
            if last_value == curr_value:
                self.unobserve([evidence.node])
            else:
                self.update_evidences([evidence])
        else:
            self.update_evidences([evidence])

        return self

    @staticmethod
    def to_dict(jt):
        """
        Converts a junction tree to a serializable dictionary.

        :param jt: Junction tree.
        :return: Dictionary.
        """

        def nodes_to_dict(nodes):
            d = {}
            for n in nodes:
                if isinstance(n, SepSet):
                    d[n.id] = {
                        'left': n.left.id,
                        'right': n.right.id,
                        'type': 'sepset'
                    }
                elif isinstance(n, Clique):
                    d[n.id] = {
                        'node_ids': list(n.node_ids),
                        'type': 'clique'
                    }
            return d

        def edges_to_dict(edges):
            return [e.sep_set.id for e in edges]

        bbn_nodes = {n.id: n.to_dict() for n in jt.get_bbn_nodes()}
        jt_nodes = nodes_to_dict(jt.get_nodes())
        jt_edges = edges_to_dict(jt.get_edges())

        return {
            'bbn_nodes': bbn_nodes,
            'jt': {
                'nodes': jt_nodes,
                'edges': jt_edges,
                'parent_info': jt.parent_info
            }
        }

    @staticmethod
    def from_dict(d):
        """
        Converts a dictionary to a junction tree.

        :param d: Dictionary.
        :return: Junction tree.
        """

        def get_variable(d):
            return Variable(d['id'], d['name'], d['values'])

        def get_bbn_node(d):
            return BbnNode(get_variable(d['variable']), d['probs'])

        def get_clique(d, bbn_nodes):
            return Clique([bbn_nodes[idx] if idx in bbn_nodes else bbn_nodes[str(idx)] for idx in d['node_ids']])

        def get_sep_set(lhs_clique, rhs_clique):
            _, lhs, rhs, intersection = lhs_clique.intersects(rhs_clique)
            return SepSet(lhs_clique, rhs_clique, lhs, rhs, intersection)

        bbn_nodes = {k: get_bbn_node(n) for k, n in d['bbn_nodes'].items()}

        cliques = [get_clique(clique, bbn_nodes)
                   for k, clique in d['jt']['nodes'].items() if clique['type'] == 'clique']
        cliques = {c.id: c for c in cliques}

        sepsets = [get_sep_set(cliques[s['left']], cliques[s['right']])
                   for k, s in d['jt']['nodes'].items() if s['type'] == 'sepset']
        sepsets = {s.id: s for s in sepsets}

        edges = [JtEdge(sepsets[e]) for e in d['jt']['edges']]

        jt = JoinTree()
        if len(edges) > 0:
            for e in edges:
                jt.add_edge(e)
        else:
            jt.nodes = cliques

        jt.parent_info = {int(k): v for k, v in d['jt']['parent_info'].items()}
        return jt

    def __shouldadd__(self, edge):
        """
        Checks if the specified edge should be added.

        :param edge: Edge.
        :return: A boolean indicating if the specified edge should be added.
        """
        lhs = edge.i
        rhs = edge.j

        if lhs.id == rhs.id:
            return False

        if not PathDetector(self, lhs.id, rhs.id).exists():
            return True

        return False

    def __get_parent_ids__(self, id):
        """
        Gets the parent ids of the specified node id.

        :param id: Node id.
        :return: Array of parent ids.
        """
        node = self.get_bbn_node(id)
        if 'parents' in node.metadata:
            return [n.id for n in node.metadata['parents']]
        return []

    def __notify_listener__(self, change):
        """
        Notifies the JoinTreeListener, if any.

        :param change: ChangeType.
        """
        if self.listener is None:
            return
        if ChangeType.RETRACTION == change:
            self.listener.evidence_retracted(self)
        elif ChangeType.UPDATE == change:
            self.listener.evidence_updated(self)


class PathDetector(object):
    """
    Detects path between two nodes.
    """

    def __init__(self, graph, start, stop):
        """
        Ctor.

        :param graph: Join tree.
        :param start: Start node id.
        :param stop: Stop node id.
        """
        self.graph = graph
        self.start = start
        self.stop = stop
        self.seen = set()

    def exists(self):
        """
        Checks if a path exists.

        :return: True if a path exists, otherwise, false.
        """
        if self.start == self.stop:
            return True
        else:
            return self.__find__(self.start)

    def __find__(self, i):
        """
        Checks if a path exists from the specified node to the stop node.

        :param i: Node id.
        :return: True if a path exists, otherwise, false.
        """
        neighbors = set()

        try:
            neighbors = self.graph.get_neighbors(i)
        except KeyError:
            pass

        if self.stop in neighbors:
            return True

        self.seen.add(i)
        for neighbor in neighbors:
            if neighbor not in self.seen and self.__find__(neighbor):
                return True

        return False


class JoinTreeListener(object):
    """
    Interface like class used for listening to a join tree.
    """

    def evidence_retracted(self, join_tree):
        """
        Evidence is retracted.

        :param join_tree: Join tree.
        """
        pass

    def evidence_updated(self, join_tree):
        """
        Evidence is updated.

        :param join_tree: Join tree.
        """
        pass


class EvidenceType(Enum):
    """
    Evidence type.
    """
    VIRTUAL = 1
    FINDING = 2
    OBSERVATION = 3
    UNOBSERVE = 4


class ChangeType(Enum):
    """
    Change type.
    """
    NONE = 1
    UPDATE = 2
    RETRACTION = 3


class EvidenceBuilder(object):
    """
    Evidence builder.
    """

    def __init__(self):
        """
        Ctor.
        """
        self.values = dict()
        self.node = None
        self.type = EvidenceType.OBSERVATION

    def with_node(self, node):
        """
        Adds a BBN node.

        :param node: BBN node.
        :return: Builder.
        """
        self.node = node
        return self

    def with_type(self, type):
        """
        Adds the EvidenceType.

        :param type: EvidenceType.
        :return: Builder.
        """
        self.type = type
        return self

    def with_evidence(self, val, likelihood):
        """
        Adds evidence.

        :param val: Value.
        :param likelihood: Likelihood.
        :return: Builder.
        """
        self.values[val] = likelihood
        return self

    def build(self):
        """
        Builds an evidence.

        :return: Evidence.
        """
        evidence = Evidence(self.node, self.type)
        for k, v in self.values.items():
            evidence.add_value(k, v)
        return evidence


class Evidence(object):
    """
    Evidence.
    """

    def __init__(self, node, type):
        """
        Ctor.

        :param node: BBN node.
        :param type: EvidenceType.
        """
        self.node = node
        self.type = type
        self.values = dict()

    def add_value(self, value, likelihood):
        """
        Adds a value.

        :param value: Value.
        :param likelihood: Likelihood.
        :return: This evidence.
        """
        self.values[value] = likelihood
        return self

    def compare(self, potentials):
        """
        Compares this evidence with previous ones.

        :param potentials: Map of potentials.
        :return: The ChangeType from the comparison.
        """
        that = self.__convert__(potentials)

        that_unobserve = self.__is_unobserved__(that)
        this_unobserve = self.__is_unobserved__(self.values)

        if that_unobserve and this_unobserve:
            return ChangeType.NONE

        that_observe = self.__is_observed__(that)
        this_observe = self.__is_observed__(self.values)

        if that_observe and this_observe:
            s1 = self.__get_observed_value__(that)
            s2 = self.__get_observed_value__(self.values)

            if s1 == s2:
                return ChangeType.NONE
            else:
                return ChangeType.RETRACTION

        return ChangeType.RETRACTION

    @staticmethod
    def __convert__(potentials):
        """
        Converts potentials to a map (dict).

        :param potentials: Potentials.
        :return: Dict where keys are BBN node values and values are likelihoods.
        """
        m = dict()
        for k, v in potentials.items():
            m[k] = v.entries[0].value
        return m

    @staticmethod
    def __is_unobserved__(values):
        """
        Checks if the values represent an unobserved evidence. If all likelihoods are 1.0, then this
        map of values represent unobserved evidence.

        :param values: Map of values, where keys are values and values are likelihoods.
        :return: A boolean indicating if the values represent unobserved evidence.
        """
        count = 0
        for k, v in values.items():
            count += v
        return count == len(values)

    @staticmethod
    def __is_observed__(values):
        """
        Checks if the values represent an observed evidence. If all likelihoods are 0 with exactly
        one of them being 1, then this map of values represent observed evidence.

        :param values: Map of values, where keys are values and values are likelihoods.
        :return: A boolean indicating if the values represent observed evidence.
        """
        one = 0
        zero = 0

        for k, v in values.items():
            if 1.0 == v:
                one += 1
            else:
                zero += 1

        return 1 == one and len(values) - 1 == zero

    @staticmethod
    def __get_observed_value__(values):
        """
        Gets the value that is observed (the value whose likelihood is 1.0).

        :param values: Map of values, where keys are values and values are likelihoods.
        :return: Observed value.
        """
        strs = [k for k in values if 1.0 == values[k]]
        return strs[0]

    @staticmethod
    def __normalize_value_between_zero_one__(val):
        """
        Normalizes the specified value to the range [0.0, 1.0]. If the specified value is less than 0.0 then
        0.0 is returned; if the specified value is greater than 1.0 then 1.0 is returned.

        :param val: Value.
        :return: A value in the range [0.0, 1.0].
        """
        if 0.0 <= val <= 1.0:
            return val
        elif val < 0.0:
            return 0.0
        else:
            return 1.0

    @staticmethod
    def __normalize_value_zero_or_one__(val):
        """
        Normalizes the specified value to either 0.0 or 1.0 (and nothing else). If the specified value is anything
        greater than 0.0, then a 1.0 is returned, else a 0.0 is returned.

        :param val: Value.
        :return: 0.0 or 1.0.
        """
        if val > 0.0:
            return 1.0
        else:
            return 0.0

    def validate(self):
        """
        Validates this evidence.

        * virtual evidence: each likelihood must be in the range [0, 1].
        * finding evidence: all likelihoods must be exactly 1.0 or 0.0.
        * observation evidence: exactly one likelihood is 1.0 and all others must be 0.0.
        """
        for value in self.node.variable.values:
            if value not in self.values:
                self.values[value] = 0.0

        if EvidenceType.VIRTUAL == self.type:
            for value in self.node.variable.values:
                self.values[value] = Evidence.__normalize_value_between_zero_one__(self.values[value])
        elif EvidenceType.FINDING == self.type:
            for value in self.node.variable.values:
                self.values[value] = Evidence.__normalize_value_zero_or_one__(self.values[value])

            count = sum([x for x in self.values.values()])
            if 0.0 == count:
                for value in self.node.variable.values:
                    self.values[value] = 1.0
        elif EvidenceType.OBSERVATION == self.type:
            tuples = []
            for k, v in self.values.items():
                pair = (k, v)
                tuples.append(pair)
            tuples = sorted(tuples, key=lambda x: (x[1]), reverse=True)
            key = tuples[0][0]
            for value in self.node.variable.values:
                if key == value:
                    self.values[value] = 1.0
                else:
                    self.values[value] = 0.0
        elif EvidenceType.UNOBSERVE == self.type:
            for value in self.node.variable.values:
                self.values[value] = 1.0



from copy import deepcopy


class Variable(object):
    """
    A variable.
    """

    def __init__(self, id, name, values):
        """
        Ctor.

        :param id: Numeric identifier. e.g. 0
        :param name: Name. e.g. 'a'
        :param values: Array of values. e.g. ['on', 'off']
        """
        self.id = id
        self.name = name
        self.values = values

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memodict={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memodict))
        return result

    def __str__(self):
        return '{}|{}|{}'.format(self.id, self.name, self.values)

    def to_dict(self):
        """
        Gets a JSON serializable dictionary representation.

        :return: Dictionary.
        """
        return {
            'id': self.id,
            'name': self.name,
            'values': self.values
        }



import itertools

import numpy as np


class DiscreteData(object):
    """
    A discrete dataset.
    """

    def __init__(self, df):
        """
        Ctor.

        :param df: Dataframe.
        """
        self.df = df
        if self.__is_valid__() is False:
            raise Exception('Data is not valid.')
        self.variable_profiles = self.get_variable_profiles()
        self.N = df.shape[0]
        self.probs = {}
        self.counts = {}

    def get_variables(self, by_name=True):
        """
        Gets a dictionary of variables. Keys could be the name or numeric id.

        :param by_name: A boolean indicating if keys should be names or not. True by default.
        :return: Dictionary.
        """
        variables = {}
        for i, col in enumerate(self.df.columns):
            if by_name is True:
                variables[col] = i
            else:
                variables[i] = col
        return variables

    def get_variable_profiles(self):
        """
        Gets a dictionary of values for each variable. Keys are variable/column names and values are list
        of the domain/values of the variable.

        :return: Dictionary.
        """
        profile = {}
        for col in self.df.columns:
            s = self.df[col].value_counts()
            profile[col] = list(s.index)
        return profile

    def __is_valid__(self):
        """
        Checks to see if the dataframe is valid. Each variable/column must have more than 1 unique (at least 2) values.

        :return: Boolean indicating if the data is valid.
        """
        valid = True
        for col in self.df.columns:
            s = self.df[col].value_counts()
            if len(s) < 2:
                valid = False
                break
        return valid

    def __sort__(self, names, vals):
        """
        Sorts the names and corresponding values ascendingly by name.

        :param names: List of names.
        :param vals: List of values.
        :return: List of tuple, where each tuple is a (name, value).
        """
        pairs = [(n, v) for n, v in zip(names, vals)]
        pairs = sorted(pairs, key=lambda t: t[0])
        return pairs

    def __to_query__(self, names, vals):
        """
        Converts the names and corresponding values to a query filter. For example, if names = ['a', 'c', 'b']
        and values = ['true', 'false', 'true'], then a filter is generated as follows, 'a == "true" and b == "true"
        and c == "false"'.

        :param names: List of names.
        :param vals: List of values.
        :return: Query filter for use with dataframe.query(...) method.
        """
        pairs = self.__sort__(names, vals)
        pairs = ' and '.join(['{} == "{}"'.format(t[0], t[1]) for t in pairs])
        return pairs

    def __count__(self, names, vals):
        """
        Counts the frequency of the combination of the specified values.

        :param names: Names of variables.
        :param vals: Values of variables.
        :return: Count.
        """
        q = self.__to_query__(names, vals)
        if q in self.counts:
            return self.counts[q]
        else:
            n = self.df.query(q).shape[0]
            self.counts[q] = n
            return self.counts[q]

    def __count_parents_child__(self, ch_name, ch_val, pa_names, pa_vals):
        """
        Counts the frequency of the specified combination of parent and child values.

        :param ch_name: Child name.
        :param ch_val: Child value.
        :param pa_names: Parent name.
        :param pa_vals: Parent value.
        :return: Count.
        """
        names = [ch_name]
        names.extend(pa_names)

        vals = [ch_val]
        vals.extend(pa_vals)

        return self.__count__(names, vals)

    def __get_prob__(self, name, val):
        """
        Gets the probability of a variable equal to the specified value.

        :param name: Name of variable/column.
        :param val: Value.
        :return: Probability, P(name = val).
        """
        q = '{} == "{}"'.format(name, val)
        if q in self.probs:
            return self.probs[q]

        n = self.df.query(q).shape[0]
        p = n / float(self.N)
        self.probs[q] = p
        return p

    def __get_joint_prob__(self, name1, val1, name2, val2):
        """
        Gets the joint probability of two variables equal to the corresponding values.

        :param name1: Name of variable/column.
        :param val1: Value of variable.
        :param name2: Name of variable/column.
        :param val2: Value of variable.
        :return: Probability, P(name1 = val1, name2 = val2).
        """
        q = self.__to_query__([name1, name2], [val1, val2])
        if q in self.probs:
            return self.probs[q]

        n = self.df.query(q).shape[0]
        p = n / float(self.N)
        self.probs[q] = p
        return p

    def __get_joint_prob_set__(self, name1, val1, cond_names, cond_vals):
        """
        Gets the joint probability of a variable given a set of variables.

        :param name1: Name of variable/column.
        :param val1: Value of variable.
        :param cond_names: List of names of variables/columns.
        :param cond_vals: List of values of variables.
        :return: Joint probability, P(name1 = val1, cond_names = cond_vals).
        """
        # compute P(x_i, x_k)
        names = [name1]
        names.extend(cond_names)

        vals = [val1]
        vals.extend(cond_vals)

        q_ik = self.__to_query__(names, vals)
        p_ik = None

        if q_ik in self.probs:
            p_ik = self.probs[q_ik]
        else:
            n = self.df.query(q_ik).shape[0]
            p = n / float(self.N)
            self.probs[q_ik] = p
            p_ik = self.probs[q_ik]

        return p_ik

    def __get_joint_prob_triplet__(self, name1, val1, name2, val2, cond_names, cond_vals):
        """
        Gets the joint probability of three variables equal to the corresponding values. Note that the last
        two parameters should be a list of variable names and values.

        :param name1: Name of variable/column.
        :param val1: Value of variable.
        :param name2: Name of variable/column.
        :param val2: Value of variable.
        :param cond_names: List of names of variables/columns.
        :param cond_vals: List of values of variables.
        :return: Probability, P(name1 = val1, name2 = val2, cond_names = cond_vals).
        """
        # compute P(x_i, x_j, x_k)
        names = [name1, name2]
        names.extend(cond_names)

        vals = [val1, val2]
        vals.extend(cond_vals)

        q_ijk = self.__to_query__(names, vals)
        p_ijk = None

        if q_ijk in self.probs:
            p_ijk = self.probs[q_ijk]
        else:
            n = self.df.query(q_ijk).shape[0]
            p = n / float(self.N)
            self.probs[q_ijk] = p
            p_ijk = self.probs[q_ijk]

        return p_ijk

    def __get_cond_prob__(self, name1, val1, name2, val2):
        """
        Gets the conditional probability of two variables, namely, P(name1 | name2).

        :param name1: Name of variable/column.
        :param val1: Value of variable.
        :param name2: Name of variable/column.
        :param val2: Value of variable.
        :return: Conditional probability, P(name1 = val1 | name2 = val2).
        """
        p_ij = self.__get_joint_prob__(name1, val1, name2, val2)
        p_j = self.__get_prob__(name2, val2)
        p_i_j = p_ij / p_j
        return p_i_j

    def __get_cond_prob_set__(self, name1, val1, cond_names, cond_vals):
        """
        Gets the conditional probability of a variable given a set of variables, namely, P(name1 | cond_names).

        :param name1: Name of variable/column.
        :param val1: Value of variable.
        :param cond_names: List of names of variables/columns.
        :param cond_vals: List of values of variables.
        :return: Conditional probability, P(name1 = val1 | cond_names = cond_vals).
        """
        # compute P(x_i, x_k)
        p_ik = self.__get_joint_prob_set__(name1, val1, cond_names, cond_vals)

        # compute P(x_k)
        q_k = self.__to_query__(cond_names, cond_vals)
        p_k = None

        if q_k in self.probs:
            p_k = self.probs[q_k]
        else:
            n = self.df.query(q_k).shape[0]
            p = n / float(self.N)
            self.probs[q_k] = p
            p_k = self.probs[q_k]

        try:
            p_i_k = p_ik / p_k
        except ZeroDivisionError:
            p_i_k = 0.0

        return p_i_k

    def __get_cond_prob_triplet__(self, name1, val1, name2, val2, cond_names, cond_vals):
        """
        Gets the conditional probability of two variables given a set of third variables, namely,
        P(name1, name2 | cond_names).

        :param name1: Name of variable/column.
        :param val1: Value of variable.
        :param name2: Name of variable/column.
        :param val2: Value of variable.
        :param cond_names: List of names of variables/columns.
        :param cond_vals: List of values of variables.
        :return: Conditional probability, P(name1 = val1, name2 = val2 | cond_names = cond_vals).
        """
        # compute P(x_i, x_j, x_k)
        p_ijk = self.__get_joint_prob_triplet__(name1, val1, name2, val2, cond_names, cond_vals)

        # compute P(x_k)
        q_k = self.__to_query__(cond_names, cond_vals)
        p_k = None

        if q_k in self.probs:
            p_k = self.probs[q_k]
        else:
            n = self.df.query(q_k).shape[0]
            p = n / float(self.N)
            self.probs[q_k] = p
            p_k = self.probs[q_k]

        try:
            p_ij_k = p_ijk / p_k
        except ZeroDivisionError:
            p_ij_k = 0.0

        return p_ij_k

    def __get_mi__(self, name1, name2):
        """
        Gets the mutual information between two variables.

        :param name1: Name of variable/column.
        :param name2: Name of variable/column.
        :return: Mutual information, I(name1, name2).
        """
        vals1 = self.variable_profiles[name1]
        vals2 = self.variable_profiles[name2]

        mi = 0.0
        for val1 in vals1:
            p_i = self.__get_prob__(name1, val1)
            for val2 in vals2:
                p_j = self.__get_prob__(name2, val2)
                p_ij = self.__get_joint_prob__(name1, val1, name2, val2)

                if p_i != 0.0 and p_j != 0.0:
                    x = p_ij / p_i / p_j

                    if x == 0.0:
                        pass
                    else:
                        weight = np.log(x)
                        mi += p_ij * weight
        return mi

    def __get_cond_mi__(self, name1, name2, cond_names):
        """
        Gets the conditional mutual information between two variables given a third set of variables.

        :param name1: Name of variable/column.
        :param name2: Name of variable/column.
        :param cond_names: List of names of variables/columns.
        :return: Conditional mutual information, I(name1, name2 | cond_names).
        """
        vals1 = self.variable_profiles[name1]
        vals2 = self.variable_profiles[name2]
        vals = list(itertools.product(*[self.variable_profiles[name] for name in cond_names]))

        mi = 0.0

        for val1 in vals1:
            for val2 in vals2:
                for cond_vals in vals:
                    p_ijk = self.__get_joint_prob_triplet__(name1, val1, name2, val2, cond_names, cond_vals)
                    p_ij_k = self.__get_cond_prob_triplet__(name1, val1, name2, val2, cond_names, cond_vals)
                    p_i_k = self.__get_cond_prob_set__(name1, val1, cond_names, cond_vals)
                    p_j_k = self.__get_cond_prob_set__(name2, val2, cond_names, cond_vals)

                    if p_i_k != 0.0 and p_j_k != 0.0:
                        x = p_ij_k / p_i_k / p_j_k

                        if x == 0.0:
                            pass
                        else:
                            weight = np.log(x)
                            mi += p_ijk * weight

        return mi

    def __factorial__(self, x):
        """
        Factorial.

        :param x: A number.
        :return: x!
        """
        if x <= 0.0:
            return 1

        # typically stirling's approximation would be (x / e)^x
        # for large values of x, this creates an overflow
        # we use 1 here, which doesn't seem to affect the sampling or probability estimations
        g = np.sqrt(2 * np.pi * x) * np.power(x / 2.718281, 1)
        return g

    def __gamma__(self, n):
        """
        Gamma.

        :param n: A number.
        :return: (n-1)!
        """
        return self.__factorial__(n - 1)

    def is_cond_dep(self, name1, name2, names):
        """
        Checks if two variables are conditionally dependent give a third set of variables.

        :param name1: Name of variable/column.
        :param name2: Name of variable/column.
        :param names: List of names of variables/columns.
        :return: A tuple, (conditional mutual information, boolean), to signal if two variables are
        conditionally dependent.
        """
        mi = self.__get_mi__(name1, name2)
        cmi = self.__get_cond_mi__(name1, name2, names)
        if cmi > mi:
            return cmi, True
        return cmi, False

    def get_pairwise_mutual_information(self):
        """
        Gets all unique pairwise mutual information.

        :return: List of tuples where each tuple is (name1, name2, mi).
        """
        mis = []
        variables = sorted(list(self.variable_profiles.keys()))
        for i in range(len(variables)):
            name1 = variables[i]
            for j in range(len(variables)):
                name2 = variables[j]
                if j > i:
                    mi = self.__get_mi__(name1, name2)
                    t = (name1, name2, mi)
                    mis.append(t)
        mis = sorted(mis, key=lambda t: t[2], reverse=True)
        return mis

    def get_local_kutato(self, child, parents):
        """
        Computes the local K2 (Kutato) score of a child and its parents.
        See https://www.cs.waikato.ac.nz/~remco/weka.bn.pdf. The score is converted
        to its log form to avoid numerical overflow. All scores should be negative,
        with scores closer to zero indicating better likelihood (higher scores
        indicate the parent-child relationship is more likely).

        :param child: Child.
        :param parents: Parent.
        :return: Log of the K2 score.
        """
        r_i = len(self.variable_profiles[child])
        gamma_r_i = self.__gamma__(r_i)

        q_i = list(itertools.product(*[self.variable_profiles[parent] for parent in parents]))

        log_score = 0.0

        for vals in q_i:
            N_ij = self.__count__(parents, vals)
            gamma_N_ij = self.__gamma__(r_i + N_ij)

            log_sum = np.log(gamma_r_i) - np.log(gamma_N_ij)

            for val in self.variable_profiles[child]:
                N_ijk = self.__count_parents_child__(child, val, parents, vals)
                factorial_N_ijk = self.__factorial__(N_ijk)
                log_sum += np.log(factorial_N_ijk)

            log_score += log_sum

        return log_score



from pybbn.graph.potential import PotentialUtil


class Initializer(object):
    """
    Initializes the join tree.
    """

    @staticmethod
    def initialize(join_tree):
        """
        Starts the initialization.

        :param join_tree: Join tree.
        :return: Join tree.
        """
        for clique in join_tree.get_cliques():
            potential = PotentialUtil.get_potential_from_nodes(clique.nodes)
            join_tree.add_potential(clique, potential)

        for sep_set in join_tree.get_sep_sets():
            potential = PotentialUtil.get_potential_from_nodes(sep_set.nodes)
            join_tree.add_potential(sep_set, potential)

        nodes = join_tree.get_bbn_nodes()
        for node in nodes:
            clique = Initializer.get_clique(node, join_tree)
            # print('{} mapped to clique {}'.format(node.variable.name, clique))
            p1 = join_tree.potentials[clique.id]
            p2 = node.potential
            # print(p1)
            # print('>>>>>')
            # print(p2)
            # print('----')
            PotentialUtil.multiply(p1, p2)
            # print(p1)
            # print('****')

        for node in nodes:
            for value in node.variable.values:
                clique = node.metadata['parent.clique']
                clique_potential = join_tree.potentials[clique.id]
                node_potential = join_tree.get_evidence(node, value)
                PotentialUtil.multiply(clique_potential, node_potential)
                # print(clique)
                # print(clique_potential)
        return join_tree

    @staticmethod
    def get_clique(node, join_tree):
        """
        Gets the parent clique associated with the specified BBN node.

        :param node: BBN node.
        :param join_tree: Join tree.
        :return: Parent clique.
        """
        if 'parent.clique' not in node.metadata:
            cliques = sorted(join_tree.find_cliques_with_node_and_parents(node.id), key=lambda x: x.id)
            clique = cliques[0]
            node.add_metadata('parent.clique', clique)
            return clique
        else:
            return node.metadata['parent.clique']



from copy import deepcopy
from functools import reduce


class Node(object):
    """
    A node.
    """

    def __init__(self, id):
        """
        Ctor.

        :param id: Numeric identifier.
        """
        self.id = id
        self.metadata = dict()

    def add_metadata(self, k, v):
        """
        Adds metadata.

        :param k: Key. Typically a string value.
        :param v: Value. Any object.
        """
        self.metadata[k] = v

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memodict={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memodict[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memodict))
        return result

    def __str__(self):
        return '{}'.format(self.id)


class BbnNode(Node):
    """
    A BBN node.
    """

    def __init__(self, variable, probs):
        """
        Ctor.

        :param variable: A variable.
        :param probs: Array of conditional probabilities.
        """
        Node.__init__(self, variable.id)
        self.variable = variable
        self.probs = probs
        self.potential = None
        self.weight = len(self.variable.values)

    def get_weight(self):
        """
        Gets the weight, which is the number of values.

        :return: Weight.
        """
        return self.weight

    def to_dict(self):
        """
        Gets a JSON serializable dictionary representation.

        :return: Dictionary.
        """
        return {
            'probs': [p for p in self.probs],
            'variable': self.variable.to_dict()
        }

    def __str__(self):
        return '{}|{}|{}'.format(self.id, self.variable.name, str.join(',', self.variable.values))


class Clique(Node):
    """
    A clique.
    """

    def __init__(self, nodes):
        """
        Ctor.

        :param nodes: An array of BbnNodes.
        """
        nids = [n.id for n in nodes]
        nids.sort()
        sid = str.join('-', [str(x) for x in nids])
        Node.__init__(self, sid)
        self.nodes = nodes
        self.marked = False
        self.node_ids = set([node.id for node in self.nodes])

    def intersects(self, that):
        """
        Gets intersection information.

        :param that: Clique.
        :return: Tuple where first item is a boolean indicating if there is any intersection, second item are the IDs in this clique, third item are the IDs of that clique and last item are IDs common to both Cliques.
        """
        lhs = [x.id for x in self.nodes]
        rhs = [x.id for x in that.nodes]
        intersection = [x for x in lhs if x in rhs]
        is_intersection = True if len(intersection) > 0 else False
        if is_intersection:
            lhs = sorted(lhs)
            rhs = sorted(rhs)
            intersection = sorted(intersection)

        return is_intersection, lhs, rhs, intersection

    def get_sid(self):
        """
        Gets the string ID of this clique.

        :return: String ID composed of the sorted corresponding variables in each node.
        """
        sids = '-'.join(sorted([n.variable.name for n in self.nodes]))
        return sids

    def is_marked(self):
        """
        Checks if this clique is marked.

        :return: A boolean indicating if the clique is marked.
        """
        return self.marked

    def mark(self):
        """
        Marks this clique.
        """
        self.marked = True

    def unmark(self):
        """
        Unmarks this clique.
        """
        self.marked = False

    def get_node_ids(self):
        """
        Gets the node IDs in this clique.

        :return: An array of numeric ids of the nodes in this clique.
        """
        return self.node_ids

    def is_superset(self, that):
        """
        Checks if this clique is a superset of that clique.

        :param that: Clique.
        :return: A boolean indicating if this clique is a superset of the clique passed in.
        """
        s1 = self.get_node_ids()
        s2 = that.get_node_ids()
        return s1.issuperset(s2)

    def get_weight(self):
        """
        Gets the weight of this clique; the weight is product of the weights of the nodes in this clique.

        :return: Weight.
        """
        return reduce(lambda x, y: x * y, [node.get_weight() for node in self.nodes])

    def contains(self, id):
        """
        Checks if this clique contains the specified ID.

        :param id: Numeric id.
        :return: A boolean indicating if the specified id exists in this clique.
        """
        result = len([n for n in self.nodes if n.id == id])
        if result > 0:
            return True
        else:
            return False

    def get_sep_set(self, that):
        """
        Creates a separation-set from this node and the one passed in. The separation-set is composed
        of the intersection of the two cliques. If this node has [0, 1, 2] and the node passed in has
        [1, 2, 3], then the separation set will be [1, 2].

        :param that: Clique.
        :return: Separation-set.
        """
        _, lhs, rhs, intersection = self.intersects(that)
        return SepSet(self, that, lhs, rhs, intersection)

    def __str__(self):
        names = sorted([node.variable.name for node in self.nodes])
        return '({})'.format(str.join(',', names))


class SepSet(Clique):
    """
    Separation-set.
    """

    def __init__(self, left, right, lhs=None, rhs=None, intersection=None):
        """
        Ctor.

        :param left: Clique.
        :param right: Clique.
        """
        if lhs is None or rhs is None or intersection is None:
            _, lhs, rhs, intersection = left.intersects(right)

        sid = '-'.join(str(x) for arr in [lhs, intersection, rhs] for x in arr)
        Node.__init__(self, sid)
        self.nodes = [x for x in left.nodes if x.id in intersection]
        self.marked = False

        self.left = left
        self.right = right
        self.is_empty_intersection = False if len(intersection) > 0 else True

    @property
    def is_empty(self):
        """
        Checks if the cliques in this separation set have an empty intersection.

        :return: A boolean indicating if there is no intersection.
        """
        return self.is_empty_intersection

    @property
    def cost(self):
        """
        Gets the cost.

        :return: The cost.
        """
        return self.get_cost()

    @property
    def mass(self):
        """
        Gets the mass.

        :return: The mass.
        """
        return self.get_mass()

    def get_cost(self):
        """
        The cost is the sum of the weights of the cliques connected to this separation-set.

        :return: Cost.
        """
        return self.left.get_weight() + self.right.get_weight()

    def get_mass(self):
        """
        The mass is the number of nodes in this separation-set.

        :return: Mass.
        """
        return len(self.nodes)

    def __str__(self):
        s = '{}'.format(str.join(',', [node.variable.name for node in self.nodes]))
        s = '|{} -- {} -- {}|'.format(self.left, s, self.right)
        return s
